import unittest
from asyncio import iscoroutinefunction
from functools import wraps
from http.cookies import SimpleCookie
from urllib.parse import urlencode

from .app import BaseApp
from .requests import Request


class TestClient:
    def __init__(self, app):
        self.app = app
        self.cookies = None

    def set_cookie(self, key, value):
        if self.cookies is None:
            self.cookies = SimpleCookie()

        self.cookies[key] = value

    async def request(self, method, path,
                      query=None, body=None, headers=None,
                      data=None):
        headers = headers or []

        if body is None:
            if data:
                body = urlencode(data)
                headers.append(
                    ('Content-Type', 'application/x-www-form-urlencoded')
                )

        if self.cookies:
            cookie_header = '; '.join(
                v.OutputString()
                for v in self.cookies.values()
            )
            headers.append(
                ('Cookie', cookie_header)
            )

        if body and isinstance(body, str):
            body = body.encode('utf-8')

        request = Request(
            app=self.app,
            reader=None,
            method=method.upper(),
            uri=path,
            headers=headers,
            body=body,
        )
        response = await self.app.process_request(request)
        if response.cookies:
            self.cookies = response.cookies

        return response

    def __getattr__(self, name):

        def _request(*argv, **kwargs):
            return self.request(name, *argv, **kwargs)

        return _request


class TestApp(BaseApp):
    pass


class WrapCoroutines(type):
    def __new__(cls, name, bases, classdict):
        for key, value in classdict.items():
            if key.startswith('test') and iscoroutinefunction(value):
                classdict[key] = _run_coroutine(value)

        return type.__new__(cls, name, bases, classdict)


class TestCase(unittest.TestCase, metaclass=WrapCoroutines):
    pass


def _run_coroutine(func):

    @wraps(func)
    def wrapper(self, *args, **kwargs):
        future = func(self, *args, **kwargs)
        self.app.loop.run_until_complete(future)
    return wrapper



from collections import namedtuple
from enum import Enum

ActionType = Enum('ActionType', ['collection', 'detail'], module=__name__)
Action = namedtuple(
    'Action',
    ['type', 'method', 'path', 'action'],
)


class ResourceActionCreator:

    def __init__(self, type):
        self.type = type

    def __getattr__(self, method):
        def create_action(path='', action=None):
            if action is None:
                action = path

            return Action(
                type=self.type,
                method=method.upper(),
                path=path,
                action=action,
            )
        return create_action


collection = ResourceActionCreator(ActionType.collection)
detail = ResourceActionCreator(ActionType.detail)



import itertools
import re
from collections import namedtuple

from .requests import Request

HEADER_ENCODING = 'ascii'  # am i right?
HTTP_LINE_SEPARATOR = b'\r\n'
START_LINE_REGEXP = re.compile(
    br'^(?P<method>\w+) (?P<uri>.+) HTTP/(?P<version>[0-9.]+)$'
)


class ParseError(ValueError):
    pass


_RawHttpRequest = namedtuple(
    'RawHttpRequest',
    ['method', 'uri', 'http_version', 'headers', 'body'],
)
_RawHttpResponse = namedtuple(
    'RawHttpResponse',
    ['http_version', 'status_code', 'reason_phrase', 'headers', 'body'],
)


class BaseParser:
    pass


class HttpParser(BaseParser):
    def __init__(self, app):
        self.app = app

    async def read_request(self, reader):
        start_line = await self._read_http_line_from_reader(reader)

        method, uri, http_version = self._parse_start_line(start_line)

        headers = []
        content_length = 0

        while True:
            line = await self._read_http_line_from_reader(reader)
            if not line:
                # header part ends
                break

            name, value = line.split(b':', 1)
            value = value.strip()
            headers.append((name, value))

            if name.lower() == b'content-length':
                content_length = int(value)

        if content_length > 0:
            body = await reader.readexactly(content_length)
        else:
            body = b''

        request = _RawHttpRequest(
            method=method,
            uri=uri,
            http_version=http_version,
            headers=headers,
            body=body,
        )
        request = self.build_request(request, reader)

        return request

    async def write_response(self, writer, response):
        response = self.build_raw_response(response)
        await self.write_raw_response(writer, response)

    async def _read_http_line_from_reader(self, reader):
        line = await reader.readuntil(HTTP_LINE_SEPARATOR)

        # remove separator
        line = line[:-len(HTTP_LINE_SEPARATOR)]

        return line

    async def _write_http_line_to_writer(self, writer, data):
        # append separator
        data = data + HTTP_LINE_SEPARATOR

        writer.write(data)

    def _parse_start_line(self, start_line):
        match = START_LINE_REGEXP.match(start_line)

        if match:
            return (
                match.group('method'),
                match.group('uri'),
                match.group('version'),
            )
        else:
            raise ParseError('can not parse start line', start_line)

    def build_request(self, raw_request, reader):
        '''
            build a requests.Request object.
            everything in raw_request are bytes.
        '''

        kwargs = {}
        for key in ('method', 'uri'):
            kwargs[key] = getattr(raw_request, key).decode(HEADER_ENCODING)

        try:
            headers = [
                (key.decode(HEADER_ENCODING), value.decode(HEADER_ENCODING))
                for key, value in raw_request.headers
            ]
        except UnicodeDecodeError:
            raise ParseError('contains non-ascii character in header')

        return Request(
            app=self.app,
            headers=headers,
            body=raw_request.body,
            reader=reader,
            **kwargs
        )

    def build_raw_response(self, response):
        '''
            build a _RawHttpResponse object.
            most of _RawHttpresponse are bytes except status_code.
        '''
        body = response.body
        if isinstance(body, str):
            body = body.encode(response.charset)

        content_length = len(body) if isinstance(body, bytes) else 0

        headers = [
            ('Content-Length', str(content_length)),
        ]

        return _RawHttpResponse(
            http_version=b'1.1',
            status_code=response.status_code,
            reason_phrase=response.status_phrase.encode(HEADER_ENCODING),
            body=body,
            headers=[
                (key.encode(HEADER_ENCODING), value.encode(HEADER_ENCODING))
                for key, value in itertools.chain(
                    headers, response.freeze_headers().items()
                )
            ],
        )

    async def write_raw_response(self, writer, response):
        await self._write_http_line_to_writer(
            writer,
            b'HTTP/%s %d %s' % (
                response.http_version,
                response.status_code,
                response.reason_phrase,
            )
        )
        for name, value in response.headers:
            await self._write_http_line_to_writer(
                writer,
                b'%s: %s' % (name, value),
            )
        await self._write_http_line_to_writer(writer, b'')
        await writer.drain()

        if isinstance(response.body, bytes):
            writer.write(response.body)
            await writer.drain()
        else:
            async for data in response.body:
                writer.write(data)
                await writer.drain()



import re
from collections import namedtuple

PatternResolveResult = namedtuple(
    'PatternResolveResult',
    ['argv', 'kwargs'],
)


class BasePattern:
    def resolve(self, params):
        '''
            test if current pattern can resolve params

            if can resolve, return a PatternResolveResult object.
            otherwise, return None.

            :param dict params: params during resolve
        '''
        if self.match(params):
            return PatternResolveResult(
                argv=(),
                kwargs={},
            )

    def match(self, params):
        '''
            test if current pattern can match params.

            return boolean.
        '''
        return False


class PathRegexpPattern(BasePattern):
    def __init__(self, regexp):
        self.regexp = re.compile(regexp)

    def resolve(self, params):
        # params['path'] should be the tail part of previous pattern
        if 'path' in params:
            path = params['path']
        else:
            path = params['request'].url.path

        match = self.regexp.match(path)

        if match:
            argv = ()
            kwargs = match.groupdict()
            if not kwargs:
                argv = match.groups()

            # path is the tail part for next pattern
            params['path'] = path[match.end():]

            return PatternResolveResult(
                argv=argv,
                kwargs=kwargs,
            )


class MethodPattern(BasePattern):
    def __init__(self, method):
        self.method = method

    def match(self, params):
        request = params['request']

        return request.method == self.method


def resolve_patterns(pattern_list, params):
    argv = []
    kwargs = {}

    for pattern in pattern_list:
        result = pattern.resolve(params)
        if not result:
            return

        argv += result.argv
        kwargs.update(result.kwargs)

    return PatternResolveResult(
        argv=argv,
        kwargs=kwargs,
    )



import cgi
import urllib.parse
from collections import namedtuple
from io import BytesIO

from .utils import QueryList

UrlStruct = namedtuple(
    'UrlStruct',
    ['scheme', 'host', 'path', 'query'],
)


class Url(UrlStruct):
    def copy(self):
        data = {}

        for name in self._fields:
            value = getattr(self, name)
            if name == 'query':
                value = value.copy()

            data[name] = value

        return self.__class__(**data)

    def replace_query(self, **kwargs):
        for key, value in kwargs.items():
            self.query.set(key, value)

        return self

    def to_uri(self):
        return '%s?%s' % (
            self.path,
            urllib.parse.urlencode(
                list(self.query.items())
            )
        )

    @classmethod
    def from_uri(cls, uri, host=None):
        parse_result = urllib.parse.urlparse(uri)

        if host is None:
            host = parse_result.netloc

        return cls(
            scheme=parse_result.scheme,
            host=host,
            path=parse_result.path,
            query=QueryList.parse_str(parse_result.query),
        )


class ContentType:
    def __init__(self, media_type, params):
        assert isinstance(params, QueryList), (
            'params must be QueryList'
        )

        self.media_type = media_type
        self.params = params

    @classmethod
    def parse_str(cls, s):
        parts = s.split(';')
        media_type = parts.pop(0)

        params = QueryList([
            tuple(p.strip().split('=', 1))
            for p in parts
        ])

        return cls(media_type, params)


def parse_multipart(body, params):
    return cgi.parse_multipart(
        BytesIO(body),
        {
            key: value.encode('ascii')
            for key, value in params.items()
        }
    )



import asyncio
import logging
import time

from .controllers import BaseController, NotFoundController
from .parsers import HttpParser, ParseError
from .responses import Response

logger = logging.getLogger('sniper.application')


class BaseApp:
    def __init__(self, urls=[], config=None,
                 parser_class=None,
                 startups=[]):
        self.urls = urls
        self.config = config or {}

        self.parser_class = parser_class or HttpParser
        self._parser = None

        self.startups = startups

        self.next_connection_id = 0

        self.loop = asyncio.get_event_loop()

    @property
    def parser(self):
        if not self._parser:
            self._parser = self.parser_class(self)

        return self._parser

    async def process_request(self, request):
        begin_time = time.time()
        connection_id = getattr(request, 'connection_id', 0)

        logger.info(
            '%d BEGIN %s %s',
            connection_id,
            request.method,
            request.raw_uri,
        )

        try:
            response = await self.get_response(request)
        except Exception as e:
            logger.exception(e)
            response = self._build_response_for_exception(request, e)

        end_time = time.time()
        elapsed_time = end_time - begin_time
        logger.info(
            '%d %s %s %d %.2fms',
            connection_id,
            request.method,
            request.raw_uri,
            response.status_code,
            elapsed_time * 1000
        )
        return response

    async def get_response(self, request):
        result = self.resolve_request(request)
        if result:
            controller_func, argv, kwargs = result
        else:
            controller_func = NotFoundController
            argv = []
            kwargs = {}

        result = controller_func(request, *argv, **kwargs)
        if isinstance(result, BaseController):
            result = result.run()

        if asyncio.iscoroutine(result):
            result = await result

        return result

    def resolve_request(self, request):
        params = {'request': request}
        for url in self.urls:
            result = url.resolve(params)
            if result:
                return result

    def _build_response_for_exception(self, request, exception):
        return Response(
            body=str(exception) + '\n',
            status_code=500,
        )


class Application(BaseApp):
    def run(self, *argv, **kwargs):
        startup_task = self.loop.create_task(self.startup(*argv, **kwargs))
        startup_task.add_done_callback(self.on_startup_done)

        try:
            self.loop.run_forever()
        except KeyboardInterrupt:
            pass

    async def startup(self, port=None, socket_path=None):
        if port:
            await asyncio.start_server(
                self._client_connected,
                port=port,
                loop=self.loop,
            )
            logger.info('server started on port %s' % (port,))
        elif socket_path:
            await asyncio.start_unix_server(
                self._client_connected,
                path=socket_path,
                loop=self.loop,
            )
            logger.info('server started on unix path %s' % (socket_path,))
        else:
            raise ValueError('one of port and socket_path must be provided.')

        for startup in self.startups:
            await startup(self.loop, self)

    def on_startup_done(self, fut):
        'exit if error when startup'
        assert fut.done(), 'startup is not done'

        err = fut.exception()
        if err:
            self.loop.stop()
            raise err

    async def _client_connected(self, reader, writer):
        connection_id = self.next_connection_id
        self.next_connection_id += 1

        logger.debug('%d connected', connection_id)
        while True:
            if reader.at_eof():
                logger.debug('%d connection closed', connection_id)
                break

            try:
                request = await self.parser.read_request(reader)
                request.connection_id = connection_id
                resp = await self.process_request(request)
                await self.parser.write_response(writer, resp)
            except asyncio.IncompleteReadError:
                writer.close()
            except Exception as e:
                logger.exception(e)

                if isinstance(e, ParseError):
                    status_code = 400
                else:
                    status_code = 500

                resp = Response(
                    body=str(e) + '\n',
                    status_code=status_code,
                )
                await self.parser.write_response(writer, resp)
                writer.close()



import datetime
from email.utils import format_datetime

from .utils import get_now, random_string


class Session:
    def __init__(self, request, engine):
        self.request = request
        self.engine = engine

        self._key = None
        self._data = None
        self._expire_at = None

        self._edited = False

    @property
    def data(self):
        if not self._is_loaded():
            self._load_session()

        return self._data

    @property
    def key(self):
        if not self._is_loaded():
            self._load_session()

        return self._key

    @property
    def expire_at(self):
        if not self._is_loaded():
            self._load_session()

        return self._expire_at

    def get(self, key, default=None):
        return self.data.get(key, default)

    def set(self, key, value):
        self.data[key] = value
        self._edited = True

    def need_save(self):
        return self._edited

    def _is_loaded(self):
        # key and expire_at may be None after load
        # so check data here
        return self._data is not None

    def _load_session(self):
        raw_data = self.engine.load(self.request)

        if raw_data:
            self._key = raw_data['key']
            self._data = raw_data['data']
            self._expire_at = raw_data['expire_at']
        else:
            self._data = {}


class BaseSessionStore:
    def __init__(self, app):
        self.app = app

    def get(self, key):
        '''
            return a dict or None
                {
                    'key': key,
                    'data': {},
                    'expire_at': datetime.datetime(
                        2017, 3, 2, 18, 27, 11, 257436
                    ),
                }
        '''
        raise NotImplementedError

    def set(self, key, data, expire_at):
        raise NotImplementedError


class SimpleSessionStore(BaseSessionStore):
    def __init__(self, *argv, **kwargs):
        super().__init__(*argv, **kwargs)

        self.data = {}

    def get(self, key):
        return self.data.get(key, None)

    def set(self, key, data, expire_at):
        self.data[key] = {
            'key': key,
            'data': data,
            'expire_at': expire_at
        }


class SessionEngine:
    COOKIE_NAME = 'sessionid'
    EXPIRE_DAYS = 7

    def __init__(self, app):
        self.app = app
        self._store = None

    def flush(self, session, response):
        if session.need_save():
            expire_at = session.expire_at
            if expire_at is None:
                expire_at = get_now() + self.get_expire_length()

            key = session.key
            if key is None:
                key = random_string(length=32)

                name = self.cookie_name
                response.cookies[name] = key
                morsel = response.cookies[name]
                morsel['expires'] = format_datetime(expire_at, usegmt=True)
                morsel['path'] = '/'
                morsel['httponly'] = True

            self.store.set(key, session.data, expire_at)

    def load(self, request):
        cookie_name = self.cookie_name
        session_key = request.cookie.get(cookie_name, None)

        if session_key:
            store_data = self.store.get(session_key)
            if store_data is None:
                return

            for attr in ['key', 'expire_at', 'data']:
                if attr not in store_data:
                    return

            if store_data['key'] != session_key:
                return

            if not self.verify_data(store_data):
                return

            return store_data

    def verify_data(self, data):
        if data['expire_at'] < get_now():
            return False

        return True

    @property
    def cookie_name(self):
        return self.COOKIE_NAME

    @property
    def store(self):
        if self._store is None:
            cls = self.app.config.get(
                'session_store_class',
                SimpleSessionStore
            )
            self._store = cls(self.app)

        return self._store

    def get_expire_length(self):
        return datetime.timedelta(days=self.EXPIRE_DAYS)


async def session_middleware(controller, get_response):
    app = controller.app

    if not hasattr(app, 'session_engine'):
        app.session_engine = SessionEngine(app)

    request = controller.request
    request.session = Session(request, app.session_engine)

    response = await get_response(controller)
    app.session_engine.flush(request.session, response)

    return response



from sniper.filters import Q


class BaseModel:
    def __init__(self, app):
        self.app = app

    def get_queryset(self):
        return []

    def get_object(self, pk):
        return None

    def get_list(self, filters=None):
        return []


class Model(BaseModel):
    def get_object(self, pk):
        objects = self.get_list(
            filters=[Q(pk=pk)]
        )
        if not objects:
            return None

        return objects[0]

    def get_list(self, filters=None):
        queryset = self.get_queryset()

        if filters:
            for f in filters:
                queryset = self.apply_filter(queryset, f)

        return queryset

    def apply_filter(self, queryset, query):
        if query.key == 'pk':
            return [i for i in queryset if i['id'] == query.value]

        return queryset



from .exceptions import HttpError
from .responses import Response


async def catch_http_errors(controller, get_response):
    try:
        return await get_response(controller)
    except HttpError as e:
        return controller.handle_http_error(e)


def build_entry(middleware_list):
    return _build_entry_by_iter(iter(middleware_list))


async def default_get_response(controller):
    return await controller.inner_run()


def _build_entry_by_iter(it):
    try:
        func = next(it)
    except StopIteration:
        return None

    get_response = _build_entry_by_iter(it) or default_get_response

    async def entry(controller):
        r = await func(
            controller,
            get_response
        )
        assert isinstance(r, Response), \
            'middleware must return a response object'
        return r

    return entry



import asyncio
import json

from . import middlewares
from .exceptions import MethodNotAllowed, NotFound
from .responses import Response
from .sessions import session_middleware
from .utils import cached_property


class BaseController:
    def __init__(self, request, *argv, **kwargs):
        self.request = request
        self.argv = argv
        self.kwargs = kwargs

        self.app = request.app

    def run(self):
        raise NotImplementedError


class Controller(BaseController):
    MIDDLEWARES = [
        middlewares.catch_http_errors,
        session_middleware,
    ]
    _middleware_entry = None   # lazy build

    response_class = Response
    authentication_classes = ()

    async def run(self):
        handler = self._get_middleware_entry()
        return await handler(self)

    async def inner_run(self):
        action = self.kwargs.get('action')

        if action:
            try:
                handler = getattr(self, action)
            except AttributeError:
                raise MethodNotAllowed()
        else:
            handler = self.handle

        self.before_handle()
        result = handler()

        if asyncio.iscoroutine(result):
            result = await result

        if not isinstance(result, Response):
            result = self.process_return_data(result)
            result = await self.create_response(result)

        return result

    @cached_property
    def user(self):
        for authenticator in self.get_authenticators():
            user = authenticator.authenticate(self.request)
            if user is not None:
                return user

    def get_authenticators(self):
        for cls in self.authentication_classes:
            yield cls(self)

    def before_handle(self):
        pass

    def handle_http_error(self, e):
        return Response(
            body=e.detail,
            status_code=e.status_code
        )

    def process_return_data(self, ret):
        return ret

    async def create_response(self, ret):
        return self.response_class(ret)

    @classmethod
    def get_middlewares(cls):
        return cls.MIDDLEWARES

    @classmethod
    def _get_middleware_entry(cls):
        if cls._middleware_entry is None:
            cls._middleware_entry = middlewares.build_entry(
                cls.get_middlewares()
            )

        return cls._middleware_entry


class NotFoundController(Controller):
    def run(self):
        return Response(
            'Not Found: %s %s\n' % (
                self.request.method,
                self.request.url.path
            ),
            status_code=404,
        )


class JsonResponseMixin:
    async def create_response(self, ret):
        return Response(
            json.dumps(ret),
            headers=[
                ('Content-Type', 'application/json'),
            ]
        )


class JsonController(JsonResponseMixin, Controller):
    pass


class MethodActionsMixin:
    def get(self):
        return self.handle()

    def post(self):
        return self.handle()


class ModelController(JsonResponseMixin, Controller):
    def get_model(self):
        return self.model_class(self.app)

    def get_pk(self):
        return int(self.kwargs['pk'])

    def retrieve(self):
        model = self.get_model()

        pk = self.get_pk()
        obj = model.get_object(pk)

        if obj is None:
            raise NotFound()

        if obj:
            return obj



from .resolvers import url, resource, include, verb, method_actions
from .actions import collection, detail


__all__ = [
    'url', 'resource', 'include', 'verb', 'method_actions',
    'collection', 'detail',
]



class BaseAuthentication:
    def __init__(self, controller):
        self.controller = controller

    def authenticate(self, request):
        '''
            return None or an object.
        '''
        pass



import json
from http.cookies import SimpleCookie
from urllib.parse import unquote

from .http import ContentType, Url, parse_multipart
from .utils import QueryList, cached_property


class Request:
    def __init__(self, app, reader, method, uri, headers=None, body=None):
        self.app = app
        self.reader = reader
        self.method = method.upper()
        self.raw_uri = uri
        self.headers = QueryList(headers or [])
        self.body = body or b''

        assert isinstance(self.body, bytes), 'body must be bytes'

        # url
        self.url = Url.from_uri(
            uri,
            host=self.headers.get('Host')
        )

    @property
    def query(self):
        return self.url.query

    @cached_property
    def content_type(self):
        s = self.headers.get('Content-Type')

        return ContentType.parse_str(s)

    @cached_property
    def cookie(self):
        cookie_header = self.headers.get('Cookie')
        return dict(
            (key, unquote(c.value))
            for key, c in SimpleCookie(cookie_header).items()
        )

    @cached_property
    def data(self):
        content_type = self.content_type
        media_type = content_type.media_type

        if media_type == 'application/json':
            return json.loads(
                self.body.decode(
                    content_type.params.get('charset', 'utf-8')
                )
            )
        elif media_type == 'application/x-www-form-urlencoded':
            return QueryList.parse_str(self.body.decode('ascii'))
        elif media_type == 'multipart/form-data':
            return parse_multipart(self.body, content_type.params)



from setuptools import find_packages, setup

with open('README.rst') as f:
    LONG_DESCRIPTION = f.read()

setup(
    name="sniper",
    version="0.0.6",
    description="an asynchronous restful web framework base on asyncio",
    long_description=LONG_DESCRIPTION,
    url="https://github.com/lexdene/py-sniper",
    license='GPLv3',
    author="Elephant Liu",
    author_email="lexdene@gmail.com",
    packages=find_packages(exclude=['tests/*']),
    platforms=['any'],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Topic :: Internet :: WWW/HTTP',
    ],
)



from http import HTTPStatus
from http.cookies import SimpleCookie

from .utils import QueryList, is_async_generator

DEFAULT_CONTENT_TYPE = 'text/plain'
DEFAULT_CHARSET = 'utf-8'


class BaseResponse:
    def __init__(self, body, headers,
                 status_code, status_phrase, cookies=None,
                 content_type=None, charset=None):
        assert is_async_generator(body) or isinstance(body, (str, bytes)), (
            'body must be str or bytes or async generator, got: %s' % (
                type(body),
            )
        )
        assert isinstance(headers, QueryList), (
            'header must be a QueryList object'
        )
        assert isinstance(status_code, int)
        assert isinstance(status_phrase, str)

        self.body = body
        self.headers = headers
        self.status_code = status_code
        self.status_phrase = status_phrase
        self.content_type = content_type or DEFAULT_CONTENT_TYPE
        self.charset = charset or DEFAULT_CHARSET

        self.cookies = SimpleCookie(cookies)

    def freeze_headers(self):
        headers = []

        contains_content_type = False
        for key, value in self.headers.items():
            headers.append((key, value))

            if key == 'Content-Type':
                contains_content_type = True

        if not contains_content_type:
            headers.append((
                'Content-Type',
                '%s; charset=%s' % (self.content_type, self.charset)
            ))

        for v in self.cookies.values():
            headers.append(('Set-Cookie', v.OutputString()))

        return QueryList(headers)


class Response(BaseResponse):
    def __init__(self, body='', headers=None,
                 status_code=200, status_phrase=None, cookies=None,
                 **kwargs):
        headers = QueryList(headers or [])

        if status_phrase is None:
            status_phrase = HTTPStatus(status_code).phrase

        super().__init__(
            body, headers,
            status_code, status_phrase,
            cookies, **kwargs
        )



import datetime
import random
import string
import urllib.parse


def is_async_generator(obj):
    attrs = [
        '__aiter__',
        '__anext__',
    ]
    return all(
        hasattr(obj, key)
        for key in attrs
    )


def cached_property(func):
    return _CachedProperty(func)


class _CachedProperty:
    def __init__(self, func):
        self.__doc__ = getattr(func, '__doc__')
        self.func = func

    @property
    def name(self):
        return self.func.__name__

    def __get__(self, obj, cls):
        if obj is None:
            return self

        value = obj.__dict__[self.name] = self.func(obj)
        return value


def merge_dict(*args):
    result = dict()
    for arg in args:
        result.update(arg)

    return result


class QueryList:
    def __init__(self, data):
        assert isinstance(data, list), 'data must be list'
        self._data = data

    def __str__(self):
        return '<QueryList %s>' % ', '.join(
            (
                '%s: %s' % (repr(key), repr(value))
                for key, value in self.items()
            )
        )

    def get(self, name, default=None):
        for key, value in self._data:
            if key == name:
                return value

        return default

    def set(self, name, value):
        for index, line in enumerate(self._data):
            k, v = line

            if k == name:
                self._data[index] = (name, value)
                return

        self._data.append((name, value))

    def setdefault(self, name, value):
        for line in self._data:
            k, v = line

            if k == name:
                return

        self._data.append((name, value))

    def getlist(self, name):
        return [
            value
            for key, value in self._data
            if key == name
        ]

    def items(self):
        return iter(self._data)

    def copy(self):
        return self.__class__(list(self._data))

    @classmethod
    def parse_str(cls, s):
        if s:
            l = urllib.parse.parse_qsl(
                s,
                keep_blank_values=True,
                strict_parsing=True
            )
        else:
            l = []
        return cls(l)

    def __contains__(self, name):
        for key, _ in self._data:
            if key == name:
                return True

        return False


_DEFAULT_ALLOWED_CHARS = string.ascii_lowercase + string.digits


def random_string(allowed_chars=_DEFAULT_ALLOWED_CHARS, length=32):
    return ''.join(random.choice(allowed_chars) for _ in range(length))


def get_now():
    return datetime.datetime.now(
        datetime.timezone.utc
    )



from collections import namedtuple
from itertools import chain

from ..utils import merge_dict
from .actions import Action, ActionType
from .patterns import MethodPattern, PathRegexpPattern, resolve_patterns

ResolveResult = namedtuple(
    'ResolveResult',
    ['controller', 'argv', 'kwargs'],
)


class UrlResolver:
    def __init__(self, pattern=None, controller=None, children=None,
                 data=None, patterns=None):
        if pattern:
            self.patterns = [pattern]
        elif patterns:
            self.patterns = patterns
        else:
            raise ValueError('no pattern is provided')

        self.controller = controller
        self.children = children
        self.data = data or {}

    def resolve(self, params):
        params = params.copy()

        pattern_resolve_result = resolve_patterns(self.patterns, params)
        if not pattern_resolve_result:
            return

        kwargs = merge_dict(pattern_resolve_result.kwargs, self.data)

        if not self.children:
            return ResolveResult(
                controller=self.controller,
                argv=pattern_resolve_result.argv,
                kwargs=kwargs,
            )

        for child in self.children:
            child_result = child.resolve(params)
            if child_result:
                if child_result.controller:
                    controller = child_result.controller
                else:
                    controller = self.controller

                return ResolveResult(
                    controller=controller,
                    argv=pattern_resolve_result.argv + child_result.argv,
                    kwargs=merge_dict(kwargs, child_result.kwargs)
                )


def resolver(patterns, controller, data):
    if isinstance(controller, (tuple, list)):
        return UrlResolver(
            patterns=patterns,
            controller=None,
            children=controller,
            data=data,
        )
    else:
        return UrlResolver(
            patterns=patterns,
            controller=controller,
            children=None,
            data=data,
        )


def include(regexp, children, controller=None, data=None):
    return UrlResolver(
        pattern=PathRegexpPattern(regexp),
        controller=controller,
        children=children,
        data=data
    )


def url(regexp=None, controller=None, method=None, data=None):
    patterns = []
    if regexp:
        patterns.append(PathRegexpPattern(regexp))

    if method:
        patterns.append(MethodPattern(method))

    if not patterns:
        raise ValueError('regexp or method must be provided')

    return resolver(
        patterns=patterns,
        controller=controller,
        data=data,
    )


def verb(method, controller=None, data=None):
    return resolver([MethodPattern(method)], controller, data=data)


def _build_default_actions():
    LIST_METHOD_ACTIONS = (
        ('GET', '', 'list'),
        ('POST', '', 'create'),
        ('GET', 'new', 'new'),
    )
    DETAIL_METHOD_ACTIONS = (
        ('GET', '', 'retrieve'),
        ('PUT', '', 'update'),
        ('PATCH', '', 'partial_update'),
        ('DELETE', '', 'destroy'),
        ('GET', 'edit', 'edit'),
        ('POST', 'update', 'update'),
        ('POST', 'destroy', 'destroy'),
    )

    return tuple(chain(
        (
            Action(
                type=ActionType.collection,
                method=method,
                path=path,
                action=action
            )
            for method, path, action in LIST_METHOD_ACTIONS
        ),
        (
            Action(
                type=ActionType.detail,
                method=method,
                path=path,
                action=action,
            )
            for method, path, action in DETAIL_METHOD_ACTIONS
        )
    ))


DEFAULT_ACTIONS = _build_default_actions()


def resource(name, controller, actions=[], children=[]):
    list_action_urls = []
    detail_action_urls = []

    for action in chain(DEFAULT_ACTIONS, actions):
        if action.path:
            regexp = r'^/%s$' % action.path
        else:
            regexp = r'^$'

        _url = url(
            regexp,
            method=action.method,
            data={'action': action.action}
        )

        if action.type == ActionType.collection:
            list_action_urls.append(_url)
        elif action.type == ActionType.detail:
            detail_action_urls.append(_url)

    base_path = '/' + name
    return include(
        r'^/' + name,
        [
            # list actions
            *list_action_urls,
            # detail actions
            include(
                r'^/(?P<pk>\w+)',
                detail_action_urls
            ),
        ],
        controller=controller,
        data={'resource_base_path': base_path}
    )


def method_actions(methods=None):
    if methods is None:
        methods = ['GET', 'POST']

    return [
        verb(m, data={'action': m.lower()})
        for m in methods
    ]



class HttpError(Exception):
    '''
    Base class for exceptions
    '''
    status_code = 500
    detail = 'An error occurred.'


class BadRequest(HttpError):
    status_code = 400
    detail = 'bad request'


class NotAuthenticated(HttpError):
    status_code = 401
    detail = 'not authenticated'


class NotFound(HttpError):
    status_code = 404
    detail = 'not found'


class MethodNotAllowed(HttpError):
    status_code = 405
    detail = 'method not allowed'
